searchData={"items":[{"type":"module","title":"Pythonx","doc":"Python interpreter embedded in Elixir.\n\nPythonx runs a Python interpreter in the same OS process as your Elixir\napplication, allowing you to evaluate Python code and conveniently\nconvert between Python and Elixir data structures.\n\nThe goal of this project is to better integrate Python workflows within\nLivebook and its usage in actual projects must be done with care due to\nPython's global interpreter lock (GIL), which prevents multiple threads from\nexecuting Python code at the same time. Consequently, calling `Pythonx`\nfrom multiple Elixir processes does not provide the concurrency you might\nexpect and thus it can be a source of bottlenecks. However, this concerns\nregular Python code. Packages with CPU-intense functionality, such as `numpy`,\nhave native implementation of many functions and invoking those releases the\nGIL. GIL is also released when waiting on I/O operations. In other words,\nif you are using this library to integrate with Python, make sure it happens\nin a single Elixir process or that its underlying libraries can deal with\nconcurrent invocation. Otherwise, prefer to use Elixir's `System.cmd/3` or\n`Port`s to manage multiple Python programs via I/O.","ref":"Pythonx.html"},{"type":"module","title":"Usage (script) - Pythonx","doc":"Add Pythonx to your dependencies:\n\n```elixir\nMix.install([\n  {:pythonx, \"~> 0.4.0\"}\n])\n```\n\nInitialize the interpreter, specifying the desired Python version\nand dependencies:\n\n```elixir\nPythonx.uv_init(\"\"\"\n[project]\nname = \"project\"\nversion = \"0.0.0\"\nrequires-python = \"==3.13.*\"\ndependencies = [\n  \"numpy==2.2.2\"\n]\n\"\"\")\n```\n\nEvaluate Python code:\n\n```elixir\n{result, globals} =\n  Pythonx.eval(\n    \"\"\"\n    y = 10\n    x + y\n    \"\"\",\n    %{\"x\" => 1}\n  )\n\nPythonx.decode(result)\n#=> 11\n\nglobals\n#=> %{\n#=>   \"x\" => #Pythonx.Object<\n#=>     1\n#=>   >,\n#=>   \"y\" => #Pythonx.Object<\n#=>     10\n#=>   >\n#=> }\n```\n\nIn a dynamic evaluation environment, such as IEx and Livebook, you\ncan also use the `~PY` sigil:\n\n```elixir\nimport Pythonx\n\nx = 1\n\n~PY\"\"\"\ny = 10\nresult = x + y\n\"\"\"\n\nresult\n#=> #Pythonx.Object<\n#=>   11\n#=> >\n\ny\n#=> #Pythonx.Object<\n#=>   10\n#=> >\n```","ref":"Pythonx.html#module-usage-script"},{"type":"module","title":"Usage (application) - Pythonx","doc":"Add Pythonx to your dependencies:\n\n```elixir\ndef deps do\n  [\n    {:pythonx, \"~> 0.4.0\"}\n  ]\nend\n```\n\nConfigure the desired Python version and dependencies in your\n`config/config.exs`:\n\n```elixir\nimport Config\n\nconfig :pythonx, :uv_init,\n  pyproject_toml: \"\"\"\n  [project]\n  name = \"project\"\n  version = \"0.0.0\"\n  requires-python = \"==3.13.*\"\n  dependencies = [\n    \"numpy==2.2.2\"\n  ]\n  \"\"\"\n```\n\nWith that, you can use `Pythonx.eval/2` and other APIs in your\napplication. The downloads will happen at compile time, and the\ninterpreter will get initialized automatically on boot. All necessary\nfiles are placed in Pythonx priv directory, so it is compatible with\nElixir releases.\n\nNote that currently the `~PY` sigil does not work as part of Mix project\ncode. This limitation is intentional, since in actual applications it\nis preferable to manage the Python globals explicitly.","ref":"Pythonx.html#module-usage-application"},{"type":"module","title":"How it works - Pythonx","doc":"[CPython](https://github.com/python/cpython) (the reference\nimplementation of the Python programming language) provides a `python`\nexecutable that runs Python code, and that is the usual interface\nthat developers use to interact with the interpreter. However, most\nthe CPython functionality is also available as a dynamically linked\nlibrary (`.so`, `.dylib` or `.dll`, depending on the platform). The\n`python` executable can be thought of as a program build on top of\nthat library.\n\nWith this design, any C/C++ application can link the Python library\nand use its API to execute Python code and interact with Python objects\non a low level. Taking this a step further, any language with C/C++\ninteroperability can interact with Python in the same manner. This\nusage of CPython is referred to as embedding Python.\n\nElixir provides C/C++ interoperability via Erlang NIFs and that is\nexactly how Pythonx embeds Python. As a result, the Python interpreter\noperates in the same OS process as the BEAM.\n\nFor more details refer to the official documentation on\n[embedding Python](https://docs.python.org/3/extending/embedding.html).","ref":"Pythonx.html#module-how-it-works"},{"type":"function","title":"Pythonx.decode/1","doc":"Decodes a Python object to a term.\n\nConverts the following Python types to the corresponding Elixir terms:\n\n  * `NoneType`\n  * `bool`\n  * `int`\n  * `float`\n  * `str`\n  * `bytes`\n  * `tuple`\n  * `list`\n  * `dict`\n  * `set`\n  * `frozenset`\n\nFor all other types `Pythonx.Object` is returned.","ref":"Pythonx.html#decode/1"},{"type":"function","title":"Examples - Pythonx.decode/1","doc":"iex> {result, %{}} = Pythonx.eval(\"(1, True, 'hello world')\", %{})\n    iex> Pythonx.decode(result)\n    {1, true, \"hello world\"}\n\n    iex> {result, %{}} = Pythonx.eval(\"print\", %{})\n    iex> Pythonx.decode(result)\n    #Pythonx.Object<\n       \n    >","ref":"Pythonx.html#decode/1-examples"},{"type":"function","title":"Pythonx.encode!/2","doc":"Encodes the given term to a Python object.\n\nEncoding can be extended to support custom data structures, see\n`Pythonx.Encoder`.","ref":"Pythonx.html#encode!/2"},{"type":"function","title":"Examples - Pythonx.encode!/2","doc":"iex> Pythonx.encode!({1, true, \"hello world\"})\n    #Pythonx.Object<\n      (1, True, b'hello world')\n    >","ref":"Pythonx.html#encode!/2-examples"},{"type":"type","title":"Pythonx.encoder/0","doc":"","ref":"Pythonx.html#t:encoder/0"},{"type":"function","title":"Pythonx.eval/3","doc":"Evaluates the Python `code`.\n\nThe `globals` argument is a map with global variables to be set for\nthe evaluation. The map keys are strings, while the values can be\nany terms and they are automatically converted to Python objects\nby calling `encode!/1`.\n\nThe function returns the evaluation result and a map with the updated\nglobal variables. Note that the result is an object only if `code`\nends with an expression, otherwise it is `nil`.\n\nIf the Python code raises an exception, `Pythonx.Error` is raised and\nthe message includes the usual Python error message with traceback.\n\nAll writes to the Python standard output are sent to caller's group\nleader, while writes to the standard error are sent to the\n`:standard_error` process. Reading from the standard input is not\nsupported and raises and error.\n\n> #### Concurrency {: .info}\n>\n> The Python interpreter has a mechanism known as global interpreter\n> lock (GIL), which prevents from multiple threads executing Python\n> code at the same time. Consequently, calling `eval/2` from multiple\n> Elixir processes does not provide the concurrency you might expect\n> and thus it can be a source of bottlenecks. However, this concerns\n> regular Python code. Packages with CPU-intense functionality, such\n> as `numpy`, have native implementation of many functions and invoking\n> those releases the GIL. GIL is also released when waiting on I/O\n> operations.","ref":"Pythonx.html#eval/3"},{"type":"function","title":"Options - Pythonx.eval/3","doc":"* `:stdout_device` - IO process to send Python stdout output to.\n    Defaults to the caller's group leader.\n\n  * `:stderr_device` - IO process to send Python stderr output to.\n    Defaults to the global `:standard_error`.","ref":"Pythonx.html#eval/3-options"},{"type":"function","title":"Examples - Pythonx.eval/3","doc":"iex> {result, globals} =\n    ...>   Pythonx.eval(\n    ...>     \"\"\"\n    ...>     y = 10\n    ...>     x + y\n    ...>     \"\"\",\n    ...>     %{\"x\" => 1}\n    ...>   )\n    iex> result\n    #Pythonx.Object<\n      11\n    >\n    iex> globals[\"x\"]\n    #Pythonx.Object<\n      1\n    >\n    iex> globals[\"y\"]\n    #Pythonx.Object<\n      10\n    >\n\nYou can carry evaluation state by passing globals from one evaluation\nto the next:\n\n    iex> {_result, globals} = Pythonx.eval(\"x = 1\", %{})\n    iex> {result, _globals} = Pythonx.eval(\"x + 1\", globals)\n    iex> result\n    #Pythonx.Object<\n      2\n    >","ref":"Pythonx.html#eval/3-examples"},{"type":"function","title":"Mutability - Pythonx.eval/3","doc":"Reassigning variables will have no effect on the given `globals`,\nthe returned globals will simply hold different objects:\n\n    iex> {_result, globals1} = Pythonx.eval(\"x = 1\", %{})\n    iex> {_result, globals2} = Pythonx.eval(\"x = 2\", globals1)\n    iex> globals1[\"x\"]\n    #Pythonx.Object<\n      1\n    >\n    iex> globals2[\"x\"]\n    #Pythonx.Object<\n      2\n    >\n\nHowever, objects in `globals` are not automatically cloned, so if\nyou explicitly mutate an object, it changes across all references:\n\n    iex> {_result, globals1} = Pythonx.eval(\"x = []\", %{})\n    iex> {_result, globals2} = Pythonx.eval(\"x.append(1)\", globals1)\n    iex> globals1[\"x\"]\n    #Pythonx.Object<\n      [1]\n    >\n    iex> globals2[\"x\"]\n    #Pythonx.Object<\n      [1]\n    >","ref":"Pythonx.html#eval/3-mutability"},{"type":"macro","title":"Pythonx.sigil_PY/2","doc":"Convenience macro for Python code evaluation.\n\nThis has all the characteristics of `eval/2`, except that globals\nare handled implicitly. This means that any Elixir variables\nreferenced in the Python code will automatically get encoded and\npassed as globals for evaluation. Similarly, any globals assigned\nin the code will result in Elixir variables being defined.\n\n> #### Compilation {: .warning}\n>\n> This macro evaluates Python code at compile time, so it requires\n> the Python interpreter to be already initialized. In practice,\n> this means that you can use this sigil in environments with\n> dynamic evaluation, such as IEx and Livebook, but not in regular\n> application code. In application code it is preferable to use\n> `eval/2` regardless, to make the globals management explicit.","ref":"Pythonx.html#sigil_PY/2"},{"type":"macro","title":"Examples - Pythonx.sigil_PY/2","doc":"iex> import Pythonx\n    iex> x = 1\n    iex> ~PY\"\"\"\n    ...> y = 10\n    ...> x + y\n    ...> \"\"\"\n    #Pythonx.Object<\n      11\n    >\n    iex> x\n    1\n    iex> y\n    #Pythonx.Object<\n      10\n    >","ref":"Pythonx.html#sigil_PY/2-examples"},{"type":"function","title":"Pythonx.uv_init/2","doc":"Installs Python and dependencies using [uv](https://docs.astral.sh/uv)\npackage manager and initializes the interpreter.\n\nThe interpreter is automatically initialized using the installed\nPython. The dependency packages are added to the module search path.\n\nExpects a string with `pyproject.toml` file content, which is used\nto configure the project environment. The config requires `project.name`\nand `project.version` fields to be set. It is also a good idea to\nspecify the Python version by setting `project.requires-python`.\n\n    Pythonx.uv_init(\"\"\"\n    [project]\n    name = \"project\"\n    version = \"0.0.0\"\n    requires-python = \"==3.13.*\"\n    \"\"\")\n\nTo install Python packages, set the `project.dependencies` field:\n\n    Pythonx.uv_init(\"\"\"\n    [project]\n    name = \"project\"\n    version = \"0.0.0\"\n    requires-python = \"==3.13.*\"\n    dependencies = [\n      \"numpy==2.2.2\"\n    ]\n    \"\"\")\n\nFor more configuration options, refer to the [uv documentation](https://docs.astral.sh/uv/concepts/projects/dependencies/).","ref":"Pythonx.html#uv_init/2"},{"type":"function","title":"Options - Pythonx.uv_init/2","doc":"* `:force` - if true, runs with empty project cache. Defaults to `false`.","ref":"Pythonx.html#uv_init/2-options"},{"type":"protocol","title":"Pythonx.Encoder","doc":"A protocol for converting data structures to Python objects.\n\nThe protocol has implementation for Elixir built-in data types.\n\nIn order to define it for other data structures, you can use `Pythonx.eval/2`\nand pass necessary information as built-in types. For example, imagine\nwe have the following struct representing a complex number:\n\n    defmodule Complex do\n      defstruct [:re, :im]\n    end\n\nThe protocol implementation could look like this:\n\n    defimpl Pythonx.Encoder, for: Complex do\n      def encode(complex, _encoder) do\n        {result, %{}} =\n          Pythonx.eval(\n            \"\"\"\n            complex(re, im)\n            \"\"\",\n            %{\"re\" => complex.re, \"im\" => complex.im}\n          )\n\n        result\n      end\n    end\n\n    Pythonx.encode!(%Complex{re: 1, im: -1})\n    #=> #Pythonx.Object<\n    #=>   (1-1j)\n    #=> >\n\nWhen dealing with more complex data structures, you will want to\nreturn an object from a Python package. In that case, it is a good\nidea to raise a clear error if the package is not installed. For\nexample, here is one possible implementation for `Explorer.DataFrame`:\n\n    defimpl Pythonx.Encoder, for: Explorer.DataFrame do\n      def encode(df, _encoder) do\n        {result, %{}} =\n          Pythonx.eval(\n            \"\"\"\n            try:\n              import polars\n              result = polars.read_ipc(ipc)\n            except ModuleNotFoundError:\n              result = None\n\n            result\n            \"\"\",\n            %{\"ipc\" => Explorer.DataFrame.dump_ipc!(df)}\n          )\n\n        case Pythonx.decode(result) do\n          %Pythonx.Object{} ->\n            result\n\n          nil ->\n            raise Protocol.UndefinedError,\n              protocol: @protocol,\n              value: df,\n              description:\n                \"cannot encode Explorer.DataFrame, because the polars Python package is not installed\"\n        end\n      end\n    end","ref":"Pythonx.Encoder.html"},{"type":"function","title":"Pythonx.Encoder.encode/2","doc":"A function invoked to encode the given term to `Pythonx.Object`.","ref":"Pythonx.Encoder.html#encode/2"},{"type":"type","title":"Pythonx.Encoder.t/0","doc":"All the types that implement this protocol.","ref":"Pythonx.Encoder.html#t:t/0"},{"type":"module","title":"Pythonx.Object","doc":"A struct holding a Python object.\n\nThis is an opaque struct used to pass Python objects around in\nElixir code.","ref":"Pythonx.Object.html"},{"type":"type","title":"Pythonx.Object.t/0","doc":"","ref":"Pythonx.Object.html#t:t/0"},{"type":"exception","title":"Pythonx.Error","doc":"An exception raised when Python raises an exception.","ref":"Pythonx.Error.html"},{"type":"type","title":"Pythonx.Error.t/0","doc":"","ref":"Pythonx.Error.html#t:t/0"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.36.1"}}